from typing import Iterable, Self, Tuple, Hashable
from abc import ABC, abstractmethod


class Relation:
    """
    Ex:
    ===
        >>> A = Relation(3)

    """

    def __init__(self, arity: int):
        self.arity = arity


class Query(ABC):
    @abstractmethod
    def free_variables(self) -> tuple[str, ...]:
        ...

    @abstractmethod
    def is_boolean(self) -> bool:
        ...


class Atom(Query):
    """
    Ex:
    ===
        >>> A = Relation(3)
        >>> query = Atom(A, "x", "y", "x")
        # A(x, y, x)

    """

    def __init__(self, relation: Relation, variables: Iterable[str]):
        self.relation = relation
        self.variables = variables

    def free_variables(self) -> Tuple[str, ...]:
        L = list()
        S = set()
        for var in self.variables:
            if var not in S:
                L.append(var)
                S.add(var)
        return tuple(L)

    def is_boolean(self) -> bool:
        return self.relation.arity == 0


class CQ(Query):
    """
    Ex:
    ===
        >>> A = Relation(3)
        >>> B = Relation(2)
        >>> query = CQ(("x", "y"), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists x, y, A(x, x, x) AND A(x, y, x) AND B(y, y)
        >>> query = CQ(("y", ), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists y, A(x, x, x) AND A(x, y, x) AND B(y, y)

    """

    def __init__(self, quantified_variable: Iterable[str], atoms: Iterable[Atom]):
        self.atoms = tuple(atoms)
        self.quantified_variable = tuple(quantified_variable)

    def is_boolean(self) -> bool:
        return bool(self.free_variables())

    def free_variables(self) -> tuple[str, ...]:
        L = []
        S = set(self.quantified_variable)
        SL = set()
        for atom in self.atoms:
            for variable in atom.variables:
                if variable not in S:
                    if variable not in SL:
                        L.append(variable)
                        SL.add(variable)
        return tuple(L)

    def is_sat(self) -> bool:
        # TODO
        ...


class UCQ(Query):
    """
    Ex:
    ===
        >>> A = Relation(3)
        >>> B = Relation(2)
        >>> query1 = CQ(("x", "y"), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists x, y, A(x, x, x) AND A(x, y, x) AND B(y, y)
        >>> query2 = CQ(("y", ), (Atome(A, "x", "x", "x"), Atom(A, "x", "y", "x"), Atom(B, "y", "y")))
        # Exists y, A(x, x, x) AND A(x, y, x) AND B(y, y)
        >>> query = UCQ(query1, query2, query1)

    """

    def __init__(self, *cqs: CQ):
        self.cqs = tuple(cqs)

    def is_boolean(self) -> bool:
        # TODO
        ...

    def free_variables(self):
        # TODO
        ...


class ConcreteRelation:
    """
    Ex:
    ===
        >>> A = Relation(2)
        >>> ConctreA = ConcreteRelation(A, ((12, 32), ('a', 'b'), ((0, 1), 2)))
    """

    def __init__(self, relation: Relation, values: Iterable[tuple[Hashable, ...]]):
        values = set(values)
        values_arity = set(map(len, values))
        assert values_arity == set((relation.arity,))
        self.values = values
        self.relation = relation


class Database:
    def __init__(self, *relations: ConcreteRelation):
        self.relations = ConcreteRelation

    def evaluate(self, q: Query):
        def _evaluate(q: Query, Context: dict[str, Hashable]):
            """
            Context should map each free variable of q over a concrete value.
            """
            # TODO
            ...

        # TODO
        ...

    @classmethod
    def canonical_database(cls, query: Query) -> Self:
        # TODO
        ...
